# SPDX-License-Identifier: EUPL-1.2
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>

""" Asynchronous wrapper for Repology API """

from repology_client.client import (
    api,
    get_packages,
    get_projects,
    resolve_package,
)

__all__ = [
    "api",
    "get_packages",
    "get_projects",
    "resolve_package",
]



# SPDX-License-Identifier: EUPL-1.2
# SPDX-FileCopyrightText: 2023-2024 Anna <cyber@sysrq.in>

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'repology-client'
author = 'Anna Vyalkova'
copyright = f'2024, {author}'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autosummary',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx_prompt',
]

try:
    import notfound.extension
    extensions.append('notfound.extension')

    notfound_urls_prefix = None
except ModuleNotFoundError:
    pass

try:
    import sphinx_sitemap
    extensions.append('sphinx_sitemap')

    sitemap_locales = [None]
    sitemap_url_scheme = '{link}'
except ModuleNotFoundError:
    pass

master_doc = 'index'
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

intersphinx_mapping = {
    'aiohttp': ('https://docs.aiohttp.org/en/stable', None),
}

autodoc_default_flags = [
    'members',
    'show-inheritance',
    'inherited-members',
    'undoc-members',
]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'aiohttp_theme'
html_permalinks_icon = '#'
html_theme_options = {
    'description': 'Asynchronous Python wrapper for Repology API',
    'canonical_url': 'https://repology-client.sysrq.in/',
}
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'searchbox.html',
    ]
}

html_static_path = ['_static']
html_title = f'{project} {release}'
html_baseurl = 'https://repology-client.sysrq.in/'



# SPDX-License-Identifier: EUPL-1.2
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>

from repology_client.exceptions.resolve import MultipleProjectsFound
from repology_client.types import (
    ResolvePackageType,
    _ResolvePkg,
)


def test_multiple_projects_exception():
    pkg = _ResolvePkg("freedos", "foo", ResolvePackageType.BINARY)
    names = {
        "foo": "https://foo.example.com",
        "neofoo": "https://neofoo.example.org",
    }
    exc = MultipleProjectsFound(pkg, names.keys())

    assert exc.pkg == pkg
    assert sorted(exc.names) == ["foo", "neofoo"]



# SPDX-License-Identifier: EUPL-1.2
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>

from collections.abc import AsyncGenerator

import aiohttp
import pytest
import pytest_asyncio


@pytest_asyncio.fixture(scope="session")
async def session() -> AsyncGenerator[aiohttp.ClientSession, None]:
    timeout = aiohttp.ClientTimeout(total=30)
    test_session = aiohttp.ClientSession(timeout=timeout)
    yield test_session
    await test_session.close()


@pytest.fixture(autouse=True)
def vcr_config():
    return {"allowed_hosts": [r"repology\.org"]}



# SPDX-License-Identifier: EUPL-1.2 AND CC-BY-SA-3.0
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>
# SPDX-FileCopyrightText: 2017 Mark Amery <markrobertamery@gmail.com>

""" Utility functions and classes. """

import asyncio
import time
from collections.abc import AsyncGenerator, Callable
from contextlib import asynccontextmanager
from typing import Any

import aiohttp

from repology_client.constants import USER_AGENT


class limit():
    """
    Based on `this StackOverflow answer
    <https://stackoverflow.com/questions/35196974/aiohttp-set-maximum-number-of-requests-per-second/62503115>`_.

    Set a limit on requests per second.
    """

    def __init__(self, calls: int, period: float):
        """
        :param calls: number of calls
        :param period: time period in seconds
        """
        self.calls: int = calls
        self.period: float = period
        self.clock: Callable[[], float] = time.monotonic
        self.last_reset: float = 0.0
        self.num_calls: int = 0

    def __call__(self, func: Callable) -> Callable:
        async def wrapper(*args: Any, **kwargs: Any) -> Callable:
            if self.num_calls >= self.calls:
                await asyncio.sleep(self.__period_remaining())

            period_remaining = self.__period_remaining()

            if period_remaining <= 0:
                self.num_calls = 0
                self.last_reset = self.clock()

            self.num_calls += 1

            return await func(*args, **kwargs)

        return wrapper

    def __period_remaining(self) -> float:
        elapsed = self.clock() - self.last_reset
        return self.period - elapsed


@asynccontextmanager
async def ensure_session(
    session: aiohttp.ClientSession | None = None
) -> AsyncGenerator[aiohttp.ClientSession, None]:
    """
    Create a new client session, if necessary, and close it on exit.

    :param session: :external+aiohttp:py:module:`aiohttp` client session
    """

    keep_session = True
    if session is None:
        timeout = aiohttp.ClientTimeout(total=30)
        headers = {"user-agent": USER_AGENT}
        session = aiohttp.ClientSession(headers=headers, timeout=timeout)
        keep_session = False

    try:
        yield session
    finally:
        if not keep_session:
            await session.close()



# SPDX-License-Identifier: EUPL-1.2
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>

"""
Exceptions that can be raised by the "Project by package name" tool.
"""

from collections.abc import Iterable

from repology_client.exceptions import RepologyException
from repology_client.types import _ResolvePkg


class PackageResolveException(RepologyException):
    """
    Base class for all exceptions related to the "Project by package name" tool.
    """

    def __init__(self, pkg: _ResolvePkg, message: str | None = None):
        if message is None:
            message = f"Exception occured while resolving {pkg!s}"
        super().__init__(message)
        self.pkg = pkg


class ProjectNotFound(PackageResolveException):
    """
    Raised if Repology was requested to get project by package name but
    responded with the "404 Not Found" HTTP code.
    """

    def __init__(self, pkg: _ResolvePkg):
        message = f"No projects found for {pkg!s}"
        super().__init__(pkg, message)


class MultipleProjectsFound(PackageResolveException):
    """
    Raised if Repology was requested to get project by package name without
    automatic ambiguity resolution and responded with multiple results.

    Instances of this exception contain all package names returned by Repology.
    """

    def __init__(self, pkg: _ResolvePkg, names: Iterable[str]):
        message = f"Multiple projects found for {pkg!s}"
        super().__init__(pkg, message)

        self.names = tuple(names)



# SPDX-License-Identifier: EUPL-1.2
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>

"""
Hardcoded constants for rapidly-changing Repology API. What could possibly go
wrong?
"""

# Library package name.
PACKAGE = "repology-client"

# Library version.
VERSION = "0.1.0"

# Library homepage.
HOMEPAGE = "https://repology-client.sysrq.in"

# Library's User-agent header
USER_AGENT = f"Mozilla/5.0 (compatible; {PACKAGE}/{VERSION}; +{HOMEPAGE})"

# Base URL for all API requests.
API_URL = "https://repology.org/api/v1"

# Base URL for the "Project by package name" tool.
TOOL_PROJECT_BY_URL = "https://repology.org/tools/project-by"

# Maximum number of projects API can return.
MAX_PROJECTS = 200

# Number of projects, starting from which you should use bulk export instead.
HARD_LIMIT = 5_000



# SPDX-License-Identifier: EUPL-1.2
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>

import uuid

import aiohttp
import pytest

import repology_client
from repology_client.exceptions import (
    EmptyResponse,
    InvalidInput,
)
from repology_client.exceptions.resolve import (
    MultipleProjectsFound,
    ProjectNotFound,
)
from repology_client.types import (
    Package,
    ResolvePackageType,
)


def check_firefox_project(packages: set[Package]):
    """
    Check data returned by the ``/api/v1/project/firefox`` endpoint.
    """

    firefox_pkg: Package | None = None
    for pkg in packages:
        if pkg.repo == "gentoo" and pkg.visiblename == "www-client/firefox":
            firefox_pkg = pkg
            break

    assert firefox_pkg is not None
    assert firefox_pkg.srcname == "www-client/firefox"
    assert firefox_pkg.summary == "Firefox Web Browser"

    assert firefox_pkg.maintainers is not None
    assert "mozilla@gentoo.org" in firefox_pkg.maintainers

    assert firefox_pkg.licenses is not None
    assert "MPL-2.0" in firefox_pkg.licenses


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_raw_api(session: aiohttp.ClientSession):
    problems = await repology_client.api("/repository/freebsd/problems",
                                         session=session)
    assert len(problems) != 0


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_get_packages_empty(session: aiohttp.ClientSession):
    with pytest.raises(InvalidInput):
        await repology_client.get_packages("", session=session)


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_get_packages_notfound(session: aiohttp.ClientSession):
    with pytest.raises(EmptyResponse):
        project = uuid.uuid5(uuid.NAMESPACE_DNS, "repology.org").hex
        await repology_client.get_packages(project, session=session)


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_get_packages(session: aiohttp.ClientSession):
    packages = await repology_client.get_packages("firefox", session=session)
    check_firefox_project(packages)


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_get_projects_simple(session: aiohttp.ClientSession):
    projects = await repology_client.get_projects(count=200, session=session)
    assert len(projects) == 200


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_get_400_projects(session: aiohttp.ClientSession):
    projects = await repology_client.get_projects(count=400, session=session)
    assert len(projects) > 200


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_get_projects_start_and_end(session: aiohttp.ClientSession):
    with pytest.warns(UserWarning):
        await repology_client.get_projects("a", "b", session=session)


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_get_projects_search_failed(session: aiohttp.ClientSession):
    with pytest.raises(EmptyResponse):
        project = uuid.uuid5(uuid.NAMESPACE_DNS, "repology.org").hex
        await repology_client.get_projects(search=project, session=session)


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_get_projects_search(session: aiohttp.ClientSession):
    projects = await repology_client.get_projects(search="firefox", session=session)
    assert "firefox" in projects


@pytest.mark.vcr
@pytest.mark.skip(reason="vcrpy doesn't record a cassette")
@pytest.mark.asyncio(scope="session")
async def test_resolve_package_notfound(session: aiohttp.ClientSession):
    with pytest.raises(ProjectNotFound):
        repo = uuid.uuid5(uuid.NAMESPACE_DNS, "invalid.domain").hex
        await repology_client.resolve_package(repo, "firefox")

    with pytest.raises(ProjectNotFound):
        project = uuid.uuid5(uuid.NAMESPACE_DNS, "repology.org").hex
        await repology_client.resolve_package("freebsd", project)


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_resolve_package_multiple(session: aiohttp.ClientSession):
    with pytest.raises(MultipleProjectsFound) as exc:
        # example from https://github.com/renovatebot/renovate/issues/11435
        await repology_client.resolve_package(
            "ubuntu_20_04", "gcc", ResolvePackageType.BINARY,
            autoresolve=False
        )
    assert sorted(exc.value.names) == ["gcc-defaults", "gcc-defaults-mipsen"]


@pytest.mark.vcr
@pytest.mark.asyncio(scope="session")
async def test_resolve_package(session: aiohttp.ClientSession):
    packages = await repology_client.resolve_package("freebsd", "www/firefox")
    check_firefox_project(packages)



# SPDX-License-Identifier: EUPL-1.2
# SPDX-FileCopyrightText: 2024 Anna <cyber@sysrq.in>

""" Asynchronous wrapper for Repology API """

import json
import warnings
from pathlib import PurePosixPath
from typing import Any

import aiohttp

from repology_client.constants import (
    API_URL,
    HARD_LIMIT,
    MAX_PROJECTS,
    TOOL_PROJECT_BY_URL,
)
from repology_client.exceptions import (
    EmptyResponse,
    InvalidInput,
)
from repology_client.exceptions.resolve import (
    MultipleProjectsFound,
    ProjectNotFound,
)
from repology_client.types import (
    Package,
    ProjectsRange,
    ResolvePackageType,
    _ResolvePkg,
)
from repology_client.utils import ensure_session, limit


@limit(calls=1, period=1.0)
async def _call(location: str, params: dict | None = None, *,
                session: aiohttp.ClientSession | None = None) -> bytes:
    """
    Do a single rate-limited request.

    :param location: URL location
    :param params: URL query string parameters
    :param session: :external+aiohttp:py:module:`aiohttp` client session

    :raises repology.exceptions.EmptyResponse: on empty response
    :raises aiohttp.ClientResponseError: on HTTP errors

    :returns: raw response
    """

    async with ensure_session(session) as aiohttp_session:
        async with aiohttp_session.get(location, params=params or {},
                                       raise_for_status=True) as response:
            data = await response.read()
            if not data:
                raise EmptyResponse

    return data


async def api(endpoint: str, params: dict | None = None, *,
              session: aiohttp.ClientSession | None = None) -> Any:
    """
    Do a single API request.

    :param endpoint: API endpoint (example: ``/projects``)
    :param params: URL query string parameters
    :param session: :external+aiohttp:py:module:`aiohttp` client session

    :raises repology.exceptions.EmptyResponse: on empty response
    :raises aiohttp.ClientResponseError: on HTTP errors
    :raises json.JSONDecodeError: on JSON decode failure

    :returns: decoded JSON response
    """

    raw_data = await _call(API_URL + endpoint, params, session=session)
    data = json.loads(raw_data)
    if not data:
        raise EmptyResponse

    return data


async def get_packages(project: str, *,
                       session: aiohttp.ClientSession | None = None) -> set[Package]:
    """
    Access the ``/api/v1/project/<project>`` endpoint to list packages for a
    single project.

    :param project: project name on Repology
    :param session: :external+aiohttp:py:module:`aiohttp` client session

    :raises repology.exceptions.EmptyResponse: on empty response
    :raises repology.exceptions.InvalidInput: if ``project`` is an empty string
    :raises aiohttp.ClientResponseError: on HTTP errors

    :returns: set of packages
    """

    if not project:
        raise InvalidInput(f"Not a valid project name: {project}")

    async with ensure_session(session) as aiohttp_session:
        endpoint = PurePosixPath("/project") / project
        data = await api(str(endpoint), session=aiohttp_session)
    return {Package(**package) for package in data}


async def get_projects(start: str = "", end: str = "", count: int = 200, *,
                       session: aiohttp.ClientSession | None = None,
                       **filters: Any) -> dict[str, set[Package]]:
    """
    Access the ``/api/v1/projects/`` endpoint to list projects.

    If both ``start`` and ``end`` are given, only ``start`` is used.

    :param start: name of the first project to start with
    :param end: name of the last project to end with
    :param count: maximum number of projects to fetch
    :param session: :external+aiohttp:py:module:`aiohttp` client session

    :raises repology.exceptions.EmptyResponse: on empty response
    :raises aiohttp.ClientResponseError: on HTTP errors

    :returns: project to packages mapping
    """

    if count > HARD_LIMIT:
        warnings.warn(f"Resetting count to {HARD_LIMIT} to prevent API abuse")
        count = HARD_LIMIT

    proj_range = ProjectsRange(start, end)
    if start and end:
        warnings.warn("The 'start..end' range format is not supported by Repology API")
        proj_range.end = ""

    result: dict[str, set[Package]] = {}
    async with ensure_session(session) as aiohttp_session:
        while True:
            endpoint = PurePosixPath("/projects")
            if proj_range:
                endpoint /= str(proj_range)

            batch = await api(f"{endpoint}/", filters, session=aiohttp_session)
            for project in batch:
                result[project] = set()
                for package in batch[project]:
                    result[project].add(Package(**package))

            if len(result) >= count:
                break
            if len(batch) == MAX_PROJECTS:
                # we probably hit API limits, so…
                # …choose lexicographically highest project as a new start
                proj_range.start = max(batch)
                # …make sure we haven't already hit the requested end
                if end and proj_range.start >= end:
                    break
                # …and drop end condition as unsupported
                proj_range.end = ""
            else:
                break

    return result


async def resolve_package(repo: str, name: str,
                          name_type: ResolvePackageType = ResolvePackageType.SOURCE,
                          *, autoresolve: bool = True,
                          session: aiohttp.ClientSession | None = None) -> set[Package]:
    """
    If you don't know how a project is named on Repology and therefore cannot
    use the :py:func:`get_packages` function, use this instead.

    This function uses the ``/tools/project-by`` utility to resolve a package
    name into ``/api/v1/project/<project>`` project information.

    If you disable autoresolve, ambigous package names will raise the
    :py:class:`MultipleProjectsFound` exception. It will however contain all
    matching project names, so you can continue.

    :param repo: repository name on Repology
    :param name: package name in the repository
    :param name_type: which name is used, "source" or "binary"
    :param autoresolve: enable automatic ambiguity resolution

    :raises repology.exceptions.resolve.MultipleProjectsFound: on ambigous
    package names when automatic resolution is disabled
    :raises repology.exceptions.resolve.ProjectNotFound: on failed resolve
    resulting in the "404 Not Found" HTTP error
    :raises aiohttp.ClientResponseError: on HTTP errors (except 404)

    :returns: set of packages
    """

    params = {
        "repo": repo,
        "name": name,
        "name_type": name_type,
        "target_page": "api_v1_project",
    }
    if not autoresolve:
        params["noautoresolve"] = "on"

    pkg = _ResolvePkg(repo, name, name_type)
    try:
        async with ensure_session(session) as aiohttp_session:
            raw_data = await _call(TOOL_PROJECT_BY_URL, params,
                                   session=aiohttp_session)
    except aiohttp.ClientResponseError as err:
        if err.status == 404:
            raise ProjectNotFound(pkg) from err
        raise

    data = json.loads(raw_data)
    if (
        isinstance(data, dict)
        and (targets := data.get("targets")) is not None
    ):
        raise MultipleProjectsFound(pkg, targets.keys())

    return {Package(**package) for package in data}
